@@ -1,5 +1,5 @@
 /* Dali Clock - a melting digital clock for PalmOS.
- * Copyright (c) 1991-2002 Jamie Zawinski <jwz@jwz.org>
+ * Copyright (c) 1991-2006 Jamie Zawinski <jwz@jwz.org>
  *
  * Permission to use, copy, modify, distribute, and sell this software and its
  * documentation for any purpose is hereby granted without fee, provided that
@@ -21,6 +21,8 @@
 #endif
 
 #include "daliclock.h"
+#include "numbers.h"
+#include "hsv.h"
 
 
 /* Make sure Err*Display* are not no-ops. */
@@ -38,24 +40,18 @@
 #define ROM_20 0x02003000
 #define ROM_30 0x03003000
 #define ROM_35 0x03503000
+#define ROM_40 0x04003000
 
-
-#define SECONDS_ONLY_MODE
+#define FG_PIXEL 253
+#define BG_PIXEL (FG_PIXEL+1)
 
 #undef countof
 #define countof(x) (sizeof((x))/sizeof(*(x)))
 
-typedef unsigned char POS;
-
 #ifndef MAX_SEGS_PER_LINE
-# define MAX_SEGS_PER_LINE 2
+# define MAX_SEGS_PER_LINE 3
 #endif
 
-struct raw_number {
-  const unsigned char *bits;
-  POS width, height;
-};
-
 struct scanline {
   POS left[MAX_SEGS_PER_LINE], right[MAX_SEGS_PER_LINE];
 };
@@ -74,6 +70,13 @@ struct daliclock_prefs {
   int ticks;
 };
 
+
+struct dali_color {
+  unsigned short r, g, b;
+  int h;
+  double s, v;
+};
+
 /* The runtime settings (some initialized from system prefs, but changable.)
  */
 struct state {
@@ -88,12 +91,14 @@ struct state {
   Boolean seconds_only_p;
 
   enum { MMDDYY, DDMMYY, YYMMDD, YYDDMM, MMYYDD, DDYYMM } date_format;
+  Boolean pen_down_p;
 
   int tick;
 
   int char_width, char_height, colon_width;
 
   FrameHandle base_frames [12];
+  FrameHandle orig_frames [6];
   FrameHandle current_frames [6];
   FrameHandle target_frames [6];
   FrameHandle clear_frame;
@@ -101,71 +106,14 @@ struct state {
   signed char current_digits [6];
   signed char target_digits [6];
 
+  UInt32 rom_version;
   Int16 win_width, win_height;
+  Boolean color_p;
+  Boolean hires_p;
+  struct dali_color fg, bg;
 };
 
 
-# include "zero0.xbm"
-# include "one0.xbm"
-# include "two0.xbm"
-# include "three0.xbm"
-# include "four0.xbm"
-# include "five0.xbm"
-# include "six0.xbm"
-# include "seven0.xbm"
-# include "eight0.xbm"
-# include "nine0.xbm"
-# include "colon0.xbm"
-# include "slash0.xbm"
-
-static struct raw_number numbers_0 [] = {
-  { zero0_bits, zero0_width, zero0_height },
-  { one0_bits, one0_width, one0_height },
-  { two0_bits, two0_width, two0_height },
-  { three0_bits, three0_width, three0_height },
-  { four0_bits, four0_width, four0_height },
-  { five0_bits, five0_width, five0_height },
-  { six0_bits, six0_width, six0_height },
-  { seven0_bits, seven0_width, seven0_height },
-  { eight0_bits, eight0_width, eight0_height },
-  { nine0_bits, nine0_width, nine0_height },
-  { colon0_bits, colon0_width, colon0_height },
-  { slash0_bits, slash0_width, slash0_height },
-  { 0, 0, 0 }
-};
-
-
-#ifdef SECONDS_ONLY_MODE
-# include "zero2.xbm"
-# include "one2.xbm"
-# include "two2.xbm"
-# include "three2.xbm"
-# include "four2.xbm"
-# include "five2.xbm"
-# include "six2.xbm"
-# include "seven2.xbm"
-# include "eight2.xbm"
-# include "nine2.xbm"
-# include "colon2.xbm"
-# include "slash2.xbm"
-
-static struct raw_number numbers_2 [] = {
-  { zero2_bits, zero2_width, zero2_height },
-  { one2_bits, one2_width, one2_height },
-  { two2_bits, two2_width, two2_height },
-  { three2_bits, three2_width, three2_height },
-  { four2_bits, four2_width, four2_height },
-  { five2_bits, five2_width, five2_height },
-  { six2_bits, six2_width, six2_height },
-  { seven2_bits, seven2_width, seven2_height },
-  { eight2_bits, eight2_width, eight2_height },
-  { nine2_bits, nine2_width, nine2_height },
-  { colon2_bits, colon2_width, colon2_height },
-  { slash2_bits, slash2_width, slash2_height },
-  { 0, 0, 0 }
-};
-#endif /* SECONDS_ONLY_MODE */
-
 
 static FrameHandle
 make_blank_frame (int width, int height)
@@ -274,13 +222,20 @@ copy_frame (struct state *state, FrameHandle from_h, FrameHandle to_h)
 static Boolean
 init_numbers (struct state *state)
 {
+  struct raw_number *raw;
   int i;
-  struct raw_number *raw = numbers_0;
 
-# ifdef SECONDS_ONLY_MODE
-  if (state->seconds_only_p)
-    raw = numbers_2;
-# endif /* SECONDS_ONLY_MODE */
+  if (!state->hires_p) {
+    if (!state->seconds_only_p)
+      raw = get_raw_number_0();
+    else
+      raw = get_raw_number_1();
+  } else {
+    if (!state->seconds_only_p)
+      raw = get_raw_number_2();
+    else
+      raw = get_raw_number_3();
+  }
 
   state->char_width  = raw[0].width;
   state->char_height = raw[0].height;
@@ -290,6 +245,14 @@ init_numbers (struct state *state)
     state->base_frames [i] =
       number_to_frame (raw[i].bits, raw[i].width, raw[i].height);
 
+  for (i = 0; i < countof(state->orig_frames); i++)
+    {
+      FrameHandle fh = make_blank_frame (state->char_width,
+                                         state->char_height);
+      if (!fh) return false;
+      state->orig_frames [i] = fh;
+    }
+
   for (i = 0; i < countof(state->current_frames); i++)
     {
       FrameHandle fh = make_blank_frame (state->char_width,
@@ -326,6 +289,7 @@ free_numbers (struct state *state)
     for (i = 0; i < countof ((x)); i++) FREEIF ((x)[i]); } while (0)
 
   FREELOOP (state->base_frames);
+  FREELOOP (state->orig_frames);
   FREELOOP (state->current_frames);
   FREELOOP (state->target_frames);
   FREEIF (state->clear_frame);
@@ -445,77 +409,113 @@ fill_target_digits (struct state *state)
 #ifdef HACK_FRAMEBUFFER
 static unsigned char *frame_buffer = 0;
 static unsigned int frame_buffer_bytes_per_line = 0;
+static unsigned int frame_buffer_depth = 0;
 
-/* Call this after each trip through the event loop (at least.) */
 static void
-cache_frame_buffer (void)
+cache_frame_buffer (struct state *state)
 {
-  static int initted_p = 0;
+  frame_buffer_bytes_per_line = 0;
+  frame_buffer = 0;
 
-  if (initted_p) return;
-  initted_p = 1;
+  if (state->rom_version < ROM_30)  /* ROM earlier than 3.0 -- assume. */
+    {
+#     ifndef hwrDisplayWidth
+#       define hwrDisplayWidth 160
+#     endif
+      frame_buffer_bytes_per_line = hwrDisplayWidth/8;
+    }
+  else
+    {
+      UInt32 w = 0, h = 0, d = 0;
+      Boolean color = false;
+      WinScreenMode (winScreenModeGet, &w, &h, &d, &color);
+      if (w > 0 && h > 0)
+        {
+          frame_buffer_depth = d;
+          if (d == 1)
+            frame_buffer_bytes_per_line = (w+7) / 8;
+          else if (d == 8)
+            frame_buffer_bytes_per_line = w;
+          else if (d == 16)
+            frame_buffer_bytes_per_line = w*2;
+          else
+            frame_buffer_bytes_per_line = w*4;
+        }
+    }
 
-  {
-    UInt32 rom_version = 0;
-    FtrGet (sysFtrCreator, sysFtrNumROMVersion, &rom_version);
+  if (frame_buffer_depth > 8)
+    frame_buffer_bytes_per_line = 0; /* #### not yet supported */
 
-    frame_buffer_bytes_per_line = 0;
-    frame_buffer = 0;
+  if (frame_buffer_bytes_per_line > 0)
+    {
+      WinHandle w = WinGetDisplayWindow();
 
-    if (rom_version < ROM_30)  /* ROM earlier than 3.0 -- assume. */
-      {
-#       ifndef hwrDisplayWidth
-#         define hwrDisplayWidth 160
-#       endif
-        frame_buffer_bytes_per_line = hwrDisplayWidth/8;
-      }
-    else
-      {
-        UInt32 w = 0, h = 0, d = 0;
-        Boolean color = false;
-        WinScreenMode (winScreenModeGet, &w, &h, &d, &color);
-        if (w > 0 && h > 0 && d == 1 && color == false)
-          frame_buffer_bytes_per_line = (w+7) / 8;
-      }
+      if (!w)
+        ;
+      else if (state->rom_version < ROM_35)
+        frame_buffer = (unsigned char *) w->displayAddrV20;
+      else
+        { /* functions from "3.5 New Feature Set" */
+          BitmapPtr b = WinGetBitmap (w);
+          frame_buffer = (unsigned char *) (b ? BmpGetBits (b) : 0);
 
-    if (frame_buffer_bytes_per_line > 0)
-      {
-        WinHandle w = WinGetDisplayWindow();
+          /* If we're running on 4.0 or newer, we can ask the bitmap
+             what its row-stride is instead of guessing, above.
+          */
+          if (b && state->rom_version >= ROM_40)
+            {
+              Coord bw, bh;
+              UInt16 rowbytes = 0;
+              BmpGetDimensions (b, &bw, &bh, &rowbytes);
+              frame_buffer_bytes_per_line = rowbytes;
 
-        if (!w)
-          ;
-        else if (rom_version < ROM_35)
-          frame_buffer = (unsigned char *) w->displayAddrV20;
-        else
-          { /* functions from "3.5 New Feature Set" */
-            BitmapPtr b = WinGetBitmap (w);
-            frame_buffer = (unsigned char *) (b ? BmpGetBits (b) : 0);
-          }
-      }
-  }
+              if (bw >= state->win_width * 2 &&
+                  bh >= state->win_height * 2)
+                state->hires_p = true;
+
+              state->win_width  = bw;
+              state->win_height = bh;
+            }
+        }
+    }
 }
 #endif /* HACK_FRAMEBUFFER */
 
 
 static void
-draw_horizontal_line (Int16 x1, Int16 x2, Int16 y, Boolean black_p)
+draw_horizontal_line (int x1, int x2, int y, int screen_width, Boolean black_p)
 {
-  if (x1 == x2)
-    return;
-  else if (x1 > x2)
+  if (x1 > screen_width) x1 = screen_width;
+  else if (x1 < 0) x1 = 0;
+
+  if (x2 > screen_width) x2 = screen_width;
+  else if (x2 < 0) x2 = 0;
+
+  if (x1 == x2) return;
+
+  if (x1 > x2)
     {
       Int16 swap = x1;
       x1 = x2;
       x2 = swap;
     }
 
+/*  if (y > 205) y = 205;*/
+
 #ifdef HACK_FRAMEBUFFER
   if (frame_buffer)
     {
       unsigned char *scanline =
-        frame_buffer + (y * frame_buffer_bytes_per_line);
+        frame_buffer + (y * (long) frame_buffer_bytes_per_line);
 
-      if (black_p)
+      /* #### note: assuming depth is 1 or 8 */
+      if (frame_buffer_depth == 8)
+        {
+          int pixel = (black_p ? FG_PIXEL : BG_PIXEL);
+          for (; x1 < x2; x1++)
+            scanline[x1] = pixel;
+        }
+      else if (black_p)
         for (; x1 < x2; x1++)
           scanline[x1>>3] |= 1 << (7 - (x1 & 7));
       else
@@ -557,6 +557,7 @@ draw_frame (struct state *state, FrameHandle fh, int x, int y)
           draw_horizontal_line (x + last_right,
                                 x + line->left [px],
                                 y + py,
+                                state->win_width,
                                 state->inverted_p);
 
           /* Draw the line of this segment.
@@ -564,6 +565,7 @@ draw_frame (struct state *state, FrameHandle fh, int x, int y)
           draw_horizontal_line (x + line->left [px],
                                 x + line->right[px],
                                 y + py,
+                                state->win_width,
                                 !state->inverted_p);
 
           last_right = line->right[px];
@@ -574,6 +576,7 @@ draw_frame (struct state *state, FrameHandle fh, int x, int y)
       draw_horizontal_line (x + last_right,
                             x + state->char_width,
                             y + py,
+                            state->win_width,
                             state->inverted_p);
     }
 
@@ -586,29 +589,9 @@ static void
 draw_background (struct state *state)
 {
   RectangleType r;
-
   r.topLeft.x = 0;
   r.topLeft.y = 0;
   WinGetWindowExtent (&r.extent.x, &r.extent.y);
-
-# ifdef HACK_FRAMEBUFFER
-  if (frame_buffer)
-    {
-      unsigned long pixel = (state->inverted_p ? ~0 : 0);
-      int n = frame_buffer_bytes_per_line * r.extent.y;
-      if (n % sizeof(unsigned long))
-        {
-          unsigned char *bc = frame_buffer;
-          while (--n >= 0) *bc++ = pixel;
-        }
-      else
-        {
-          unsigned long *bi = (unsigned long *) frame_buffer;
-          while ((n -= sizeof(unsigned long)) >= 0) *bi++ = pixel;
-        }
-      }
-  else
-# endif /* HACK_FRAMEBUFFER */
   if (state->inverted_p)
     WinDrawRectangle (&r, 0);
   else
@@ -617,18 +600,23 @@ draw_background (struct state *state)
 
 
 static void draw_clock (struct state *state, Boolean colonic_p);
+static void cycle_colors (struct state *, Boolean init_p);
 
 static void
 start_sequence (struct state *state)
 {
   int i;
 
+  if (state->display_date == DDate &&
+      !state->pen_down_p)
+    state->display_date = DDateOut;
+
   for (i = 0; i < countof (state->current_frames); i++)
     {
       /* Copy the (old) target_frames into the (new) current_frames,
          since that's what's on the screen now.
        */
-      copy_frame (state, state->target_frames[i], state->current_frames[i]);
+      copy_frame (state, state->current_frames[i], state->orig_frames[i]);
     }
 
   draw_clock (state, true);
@@ -644,36 +632,49 @@ start_sequence (struct state *state)
                   (j < 0 ? state->clear_frame : state->base_frames [j]),
                   state->target_frames[i]);
     }
+
+  cycle_colors (state, false);
 }
 
 
 static void
 one_step (struct state *state,
-          FrameHandle current_h, FrameHandle target_h,
+          FrameHandle orig_h,
+          FrameHandle current_h, 
+          FrameHandle target_h,
           int tick)
 {
+  struct frame *orig_frame    = (struct frame *) MemHandleLock (orig_h);
   struct frame *current_frame = (struct frame *) MemHandleLock (current_h);
   struct frame *target_frame  = (struct frame *) MemHandleLock (target_h);
 
-  struct scanline *line  = &current_frame->scanlines [0];
-  struct scanline *target = &target_frame->scanlines [0];
+  struct scanline *orig   =    &orig_frame->scanlines [0];
+  struct scanline *curr   = &current_frame->scanlines [0];
+  struct scanline *target =  &target_frame->scanlines [0];
   int i = 0, x;
+  int ticks = state->ticks;
+  int tt = ticks+1-tick;
+
   for (i = 0; i < state->char_height; i++)
     {
 # define STEP(field) \
-         (line->field += ((int) (target->field - line->field)) / tick)
+         (curr->field = (orig->field \
+                         + (((int) (target->field - orig->field)) \
+                            * tt / ticks)))
 
       for (x = 0; x < MAX_SEGS_PER_LINE; x++)
         {
           STEP (left [x]);
           STEP (right[x]);
         }
-      line++;
+      orig++;
+      curr++;
       target++;
 # undef STEP
     }
-  MemHandleUnlock (target_h);
+  MemHandleUnlock (orig_h);
   MemHandleUnlock (current_h);
+  MemHandleUnlock (target_h);
 }
 
 
@@ -686,6 +687,7 @@ tick_sequence (struct state *state)
   for (i = 0; i < countof (state->current_frames); i++)
     if (state->target_digits[i] != state->current_digits[i])
       one_step (state,
+                state->orig_frames[i],
                 state->current_frames[i],
                 state->target_frames[i],
                 state->tick);
@@ -698,6 +700,8 @@ tick_sequence (struct state *state)
       start_sequence (state);
       state->tick = state->ticks;
     }
+
+  /* cycle_colors (state, false); */
 }
 
 
@@ -705,21 +709,17 @@ static void
 draw_clock (struct state *state, Boolean colonic_p)
 {
   int x, y, i;
-  int width = state->win_width;
-  int height = state->win_height;
-
-# ifdef HACK_FRAMEBUFFER
-  cache_frame_buffer ();
-# endif /* HACK_FRAMEBUFFER */
+  int width, height;
 
+  width = state->win_width;
+  height = state->win_height;
   x = (width - ((state->char_width * 6) + (state->colon_width * 2))) / 2;
   y = (height - state->char_height) / 2;
 
   if (state->seconds_only_p)
-    {
-      x = (width - (state->char_width * 2)) / 2;
-      x += 5; /* #### kludge */
-    }
+    x = (width - (state->char_width * 2)) / 2;
+
+  if (x < 0) x = 0;
 
 # define DIGIT(n) \
   if (state->target_digits[n] != state->current_digits[n]) \
@@ -757,6 +757,73 @@ draw_clock (struct state *state, Boolean colonic_p)
 }
 
 
+static void
+set_colors (struct state *state)
+{
+  if (!state->color_p) return;
+
+  if (state->rom_version < ROM_35)
+    return;
+
+  state->fg.r = 0x0000;
+  state->fg.g = 0x4444;
+  state->fg.b = 0x6666;
+
+  state->bg.r = 0xFFFF;
+  state->bg.g = 0x0000;
+  state->bg.b = 0x0000;
+
+  rgb_to_hsv ( state->fg.r,  state->fg.g,  state->fg.b,
+              &state->fg.h, &state->fg.s, &state->fg.v);
+  rgb_to_hsv ( state->bg.r,  state->bg.g,  state->bg.b,
+              &state->bg.h, &state->bg.s, &state->bg.v);
+
+  cycle_colors (state, true);
+}
+
+
+static void
+cycle_colors (struct state *state, Boolean init_p)
+{
+  RGBColorType table[2];
+  int hue_tick = 2;
+
+  if (!state->color_p) return;
+
+  if (state->rom_version < ROM_35)
+    return;
+
+  if (init_p) 
+    {
+      UInt32 seed = TimGetTicks();
+      hue_tick = SysRandom(seed);
+    }
+
+  state->fg.h = (state->fg.h + hue_tick) % 360;
+  state->bg.h = (state->bg.h + hue_tick + 1) % 360;  /* different rates */
+
+  hsv_to_rgb ( state->fg.h,  state->fg.s,  state->fg.v,
+              &state->fg.r, &state->fg.g, &state->fg.b);
+  hsv_to_rgb ( state->bg.h,  state->bg.s,  state->bg.v,
+              &state->bg.r, &state->bg.g, &state->bg.b);
+
+  table[0].index = FG_PIXEL;
+  table[0].r = state->fg.r >> 8;
+  table[0].g = state->fg.g >> 8;
+  table[0].b = state->fg.b >> 8;
+
+  table[1].index = BG_PIXEL;
+  table[1].r = state->bg.r >> 8;
+  table[1].g = state->bg.g >> 8;
+  table[1].b = state->bg.b >> 8;
+
+  WinPalette (winPaletteSet, FG_PIXEL, 2, table);
+
+  WinSetForeColor (FG_PIXEL);   /* probably redundant? */
+  WinSetBackColor (BG_PIXEL);
+}
+
+
 static Boolean
 do_settings_form (struct state *state, FormPtr form)
 {
@@ -808,9 +875,7 @@ do_settings_form (struct state *state, FormPtr form)
   light = FrmGetObjectPtr (form, FrmGetObjectIndex(form,SettingsButtonLight));
   c12   = FrmGetObjectPtr (form, FrmGetObjectIndex(form,SettingsButton12));
   c24   = FrmGetObjectPtr (form, FrmGetObjectIndex(form,SettingsButton24));
-# ifdef SECONDS_ONLY_MODE
   secs  = FrmGetObjectPtr (form, FrmGetObjectIndex(form,SettingsSecondsOnly));
-# endif /* SECONDS_ONLY_MODE */
   mdy   = FrmGetObjectPtr (form, FrmGetObjectIndex(form,SettingsMMDDYY));
   dmy   = FrmGetObjectPtr (form, FrmGetObjectIndex(form,SettingsDDMMYY));
   ymd   = FrmGetObjectPtr (form, FrmGetObjectIndex(form,SettingsYYMMDD));
@@ -835,7 +900,6 @@ do_settings_form (struct state *state, FormPtr form)
 
   state->inverted_p = CtlGetValue (dark);
 
-# ifdef SECONDS_ONLY_MODE
   {
     Boolean old = state->seconds_only_p;
     state->seconds_only_p = CtlGetValue (secs);
@@ -850,7 +914,6 @@ do_settings_form (struct state *state, FormPtr form)
           }
       }
   }
-# endif /* SECONDS_ONLY_MODE */
 
   if (!state->seconds_only_p)
     state->twelve_hour_time_p = CtlGetValue (c12);
@@ -891,6 +954,8 @@ do_settings_form (struct state *state, FormPtr form)
 #endif /* !HAVE_PREFSETAPPPREFERENCES_20 */
   }
 
+  set_colors (state);
+
   return true;
 }
 
@@ -916,13 +981,13 @@ app_handle_event (struct state *state, EventPtr event)
   switch (event->eType)
     {
     case menuEvent:
+      state->redraw_p = true;
       switch (event->data.menu.itemID)
         {
         case MainMenuAbout:
           form = FrmInitForm (AboutForm);
           FrmDoDialog (form);
           FrmDeleteForm (form);
-          state->redraw_p = true;
           return true;
 
         case MainMenuSettings:
@@ -946,22 +1011,27 @@ app_handle_event (struct state *state, EventPtr event)
           state->test_hack = event->data.keyDown.chr;
           return true;
         }
+      else if (KeyCurrentState() & keyBitsAll)
+        {
+          /* Display date of any of the hardware keys (arrows, center-button)
+             are pressed or held down. */
+          if (!state->seconds_only_p)
+            state->display_date = DDateIn;
+          return true;
+        }
       else
         return false;
       break;
 
-    case penDownEvent:
+    case penDownEvent:		    /* Display date while pen down */
+      state->pen_down_p = true;
       if (!state->seconds_only_p)
         state->display_date = DDateIn;
       return true;
       break;
 
-    case penUpEvent:
-      if (state->display_date == DDate) /* turn off faster if already up */
-        state->display_date = DDash;
-      else if (state->display_date != DTime)
-        state->display_date = DDateOut;
-
+    case penUpEvent:		    /* leave date ~2 seconds after pen up */
+      state->pen_down_p = false;
       return true;
       break;
 
@@ -977,6 +1047,7 @@ app_handle_event (struct state *state, EventPtr event)
             state->redraw_p = false;
             for (i = 0; i < countof(state->target_digits); i++)
               state->target_digits[i] = state->current_digits[i] = -2;
+            set_colors (state);
             draw_background (state);
           }
 
@@ -1003,6 +1074,8 @@ initialize (struct state *state)
 
   MemSet (state, sizeof(*state), 0);
 
+  FtrGet (sysFtrCreator, sysFtrNumROMVersion, &state->rom_version);
+
   PrefGetPreferences (&sysPrefs);
 
   state->twelve_hour_time_p = !Use24HourFormat (sysPrefs.timeFormat);
@@ -1064,21 +1137,21 @@ initialize (struct state *state)
   state->test_hack = 0;
   state->interval = SysTicksPerSecond() / state->ticks;
   state->tick = state->ticks;
-
-#ifndef SECONDS_ONLY_MODE
   state->seconds_only_p = 0;
-#endif
+
+  WinGetWindowExtent (&state->win_width, &state->win_height);
+
+# ifdef HACK_FRAMEBUFFER
+  cache_frame_buffer (state);
+# endif /* HACK_FRAMEBUFFER */
 
   if (!init_numbers (state))
     return false;
 
-  WinGetWindowExtent (&state->win_width, &state->win_height);
   return true;
 }
 
 
-
-
 static Boolean
 start_app (struct state *state)
 {
@@ -1097,6 +1170,14 @@ start_app (struct state *state)
   form = FrmInitForm (MainForm);
   FrmSetActiveForm (form);
   FrmDrawForm (form);
+
+  if (state->rom_version < ROM_30)
+    state->color_p = false;
+  else
+    WinScreenMode (winScreenModeGet, NULL, NULL, NULL, &state->color_p);
+
+  set_colors (state);
+
   return false;
 }
 
@@ -1129,7 +1210,8 @@ event_loop (struct state *state)
     if (!SysHandleEvent (&event))
       if (!MenuHandleEvent (NULL, &event, &error))
         if (!app_handle_event (state, &event))
-          /*FrmDispatchEvent (&event)*/; 
+          if (state->rom_version >= ROM_35)
+            FrmDispatchEvent (&event);
 
     stop = TimGetTicks();		/* stop timing event execution */
 
